#!BPY

"""
Name: 'OGRE for Torchlight (*.MESH)'
Blender: 249b
Group: 'Import'
Tooltip: 'Import Torchlight OGRE files'
	
Author: Rene Lacson
Modified: Dusho
"""

from Blender import *
from xml.dom import minidom
import bpy
import math
import os

# SETTINGS
DEBUG = False
# command line parameter to execute OgreXMLConverter
ogreXMLConverter = "D:\stuff\Torchlight_modding\orge_tools\OgreXmlConverter.exe -q"
# OGRE animation data is stored as seconds so we need a frames-per-second value
fps = 30
# OGRE uses +Y to denote up so we have to adjust it to +Z (Blender convention).
# Set to False for debugging purposes only
fix_up_axis = True
# Blender's bone.tail defaults to +Y axis relative to bone.head.
# Torchlight however uses +X for bone.tail direction so set to True to make adjustments
# Set to False if importing OGRE meshes (at least one's produced by Blender's OGRE exporter)
fix_Torchlight = True
fix_TLAnimation = True

# GLOBALS
bone_index = {}
has_skeleton = False
arm = None
hierarchy = {}
bone_root = None

bind_LocQuatInfo = {}
action_LocQuatInfo = {}


class BoneLocQuat:
	def __init__(self, Lc, Qt, Id = False):			
		self.Loc = Lc
		self.Qtr = Qt
		self.isIdentity = Id
		#print(Qt)		
		"""
		# try to normalize axis only
		vec = Mathutils.Vector(Qt.x, Qt.y, Qt.z)
		vec = vec.normalize()
		self.Qtr[0] = vec[0]
		self.Qtr[1] = vec[1]
		self.Qtr[2] = vec[2]	
		"""
		
def MultiplyQuats(q1, q2):
	q = Mathutils.Quaternion(0,0,0,0)
	
	#q1 = q1.normalize()
	#q2 = q2.normalize()
	
	#(Q1 * Q2).x = (w1x2 + x1w2 + y1z2 - z1y2)
	q.x = q1.w * q2.x + q1.x * q2.w + q1.y * q2.z - q1.z * q2.y
	#(Q1 * Q2).y = (w1y2 - x1z2 + y1w2 + z1x2)
	q.y = q1.w * q2.y - q1.x * q2.z + q1.y * q2.w + q1.z * q2.x
	#(Q1 * Q2).z = (w1z2 + x1y2 - y1x2 + z1w2)
	q.z = q1.w * q2.z + q1.x * q2.y - q1.y * q2.x + q1.z * q2.w
	#(Q1 * Q2).w = (w1w2 - x1x2 - y1y2 - z1z2)
	q.w = q1.w * q2.w - q1.x * q2.x - q1.y * q2.y - q1.z * q2.z
	
	return q
		
def BoneInfoDifference(actionInfo, bindInfo):
	loc = [actionInfo.Loc[0] - bindInfo.Loc[0], actionInfo.Loc[1] - bindInfo.Loc[1], actionInfo.Loc[2] - bindInfo.Loc[2]]
	#qtr = Mathutils.DifferenceQuats(actionInfo.Qtr, bindInfo.Qtr)
	#qtr = actionInfo.Qtr - bindInfo.Qtr
	#qtr = bindInfo.Qtr - actionInfo.Qtr
	#dstBone->getInitialOrientation().Inverse() * srcBone->getInitialOrientation(); // src is action, dest is bind
	#qtr = bindInfo.Qtr.inverse() * actionInfo.Qtr
	#qtr = MultiplyQuats(bindInfo.Qtr.normalize().inverse(), actionInfo.Qtr.normalize())
	#qtr = MultiplyQuats(bindInfo.Qtr.inverse(), actionInfo.Qtr)
	#print(qtr)
	
	#qtr = MultiplyQuats(actionInfo.Qtr, bindInfo.Qtr.inverse())
	#qtr = MultiplyQuats(bindInfo.Qtr, actionInfo.Qtr.inverse())
	
	qtr = MultiplyQuats(actionInfo.Qtr.inverse(), bindInfo.Qtr)
	
	if((math.fabs(loc[0]) < 0.001) and (math.fabs(loc[1]) < 0.001) and (math.fabs(loc[2]) < 0.001) and (math.fabs(qtr.angle) < 0.001)):
		isIdentity = True
	else:
		isIdentity = False;
	
	#print(isIdentity)
	return BoneLocQuat(loc, qtr, isIdentity)

def GetBoneDifference(name):
	return BoneInfoDifference(action_LocQuatInfo[name], bind_LocQuatInfo[name])

   
def ParseActionSkeleton(skel_xml_doc):
	global action_LocQuatInfo
	global bone_root
	
	# gather all bone data; set their position and rotations in bone_data
	for contents in skel_xml_doc.getElementsByTagName("bones"):
		for element in contents.childNodes:
			if element.localName == "bone":
				name = element.getAttribute("name")		
				loc = None
				rot = None				
				for bone_xml_data in element.childNodes:
					if bone_xml_data.localName == "position":
						x = float(bone_xml_data.getAttribute("x"))
						y = float(bone_xml_data.getAttribute("y"))
						z = float(bone_xml_data.getAttribute("z"))
						loc = [x, y, z]
					elif bone_xml_data.localName == "rotation":
						# convert angle because it is in radians and Blender Quaternions are in degrees
						angle = math.degrees(float(bone_xml_data.getAttribute("angle")))
						for axis in bone_xml_data.childNodes:
							if axis.localName == "axis":
								if name == bone_root:
									if fix_up_axis:
										quat = Mathutils.Euler(0.0, 0.0, 0.0).toQuat()
										rot = Mathutils.Euler(0.0, 0.0, 0.0).toMatrix()
										print(name)
									else:
										x = float(axis.getAttribute("x"))
										y = float(axis.getAttribute("y"))
										z = float(axis.getAttribute("z"))
										quat = Mathutils.Quaternion([x, y, z], angle)
										rot = quat.toMatrix()								
								else:
									x = float(axis.getAttribute("x"))
									y = float(axis.getAttribute("y"))
									z = float(axis.getAttribute("z"))
									quat = Mathutils.Quaternion([x, y, z], angle)
									rot = quat.toMatrix()
											
				action_LocQuatInfo[name] = BoneLocQuat(loc, quat)
				print("storing actionBone: " + name)
			
	
		
def BlenderOrientedCoords(coords):
	"""
	OGRE uses Y for altitude and Z for depth.
	Blender uses other convention so flip Y and Z.
	Set Y to negative so that front of object faces camera in orthogonal front view
	"""
	if fix_up_axis:
		retval = [coords[0], -coords[2], coords[1]]
	else:
		retval = [coords[0], coords[1], coords[2]]
	return retval



def BlenderOrientedQuaternion(x, y, z, angle):
	"""
	Torchlight's bones point in the +X axis but Blender defaults to the +Y axis direction
	Adjust if necessary to Blender's convention
	"""
	if fix_Torchlight:
		retval = Mathutils.Quaternion([-y, x, z], angle)
	else:
		retval = Mathutils.Quaternion([x, y, z], angle)
	return retval



def OpenFile(filename):
	xml_file = open(filename)
	try:
		xml_doc = minidom.parse(xml_file)
		output = xml_doc
	except:
		print "File not valid!"
		output = 'None'
	xml_file.close()
	return output

def NewBone(arm, name, hierarchy):
	new_bone = Armature.Editbone()
	new_bone.name = name
		
	# set the default bone length to 0.2 Blender units, exact lenght will be updated later
	new_bone.tail = Mathutils.Vector(0, 0, 0.2)
	
	obj = Object.Get(name)
	new_bone.matrix = obj.getMatrix()

	if fix_Torchlight:
		# make the bone point in the x-axis cause it looks better than the default y-axis direction
		temp_matrix = obj.getMatrix().rotationPart()
		new_bone.matrix = Mathutils.Matrix(temp_matrix[1], temp_matrix[0], temp_matrix[2])

	parent = obj.getParent()
	if parent != None:
		new_bone.parent = arm.bones[parent.getName()]
#		new_bone.options = [Armature.CONNECTED]

	return new_bone

	
def CreateBones(bone_root, hierarchy):

	global arm

	scene = bpy.data.scenes.active

	# create an empty armature and link it to a Blender object
	arm = Armature.Armature("Armature")
	#arm.drawType = Armature.STICK
	obj = scene.objects.new(arm, "Armature")
	# enter armature edit mode
	arm.makeEditable()

	name = bone_root
	arm.bones[name] = NewBone(arm, name, hierarchy)
	children = hierarchy[name]

	while len(children) > 0:
		name = children.pop(0)
		arm.bones[name] = NewBone(arm, name, hierarchy)
		if name in hierarchy:
			children.extend(hierarchy[name])
	
	arm.update()
	# now fix the bone lenghts
	SetupTailLengths(arm)
	
def SetupTailLengths(arm):
	arm.makeEditable()
		
	for bone in arm.bones.values():
		if(bone.hasParent()):
			bone.parent.tail = bone.head	
	
	arm.update()


def CreateSkeleton(xml_doc, folder, name):

	global bone_index
	global has_skeleton
	global hierarchy
	global bone_root
	global bind_LocQuatInfo

	scene = bpy.data.scenes.active
	
	# get the skeleton link of the mesh
	skeleton_link = xml_doc.getElementsByTagName("skeletonlink")[0]
	filename = os.path.join(folder, skeleton_link.getAttribute("name"))
	skel_xml_doc = OpenFile(filename + ".xml")
	if skel_xml_doc == "None":
		print "%s file not found!" % filename
		return

	has_skeleton = True

	# create all bones based on the bone data.
	# all bones must be created before we can assign parent/child relationship
	# create empties to mark the location and rotations of the joints first
	# also fill up bone_index for vertex weight later
	for contents in skel_xml_doc.getElementsByTagName("bones"):
		for element in contents.childNodes:
			if element.localName == "bone":
				obj = scene.objects.new("Empty", element.getAttribute("name"))
				bone_index[element.getAttribute("id")] = element.getAttribute("name")

	hierarchy = {}				
	# set parent/child relation based on OGRE xml data
	for contents in skel_xml_doc.getElementsByTagName("bonehierarchy"):
		for element in contents.childNodes:
			if element.localName == "boneparent":
				obj = Object.Get(element.getAttribute("parent"))
				obj.makeParent([Object.Get(element.getAttribute("bone"))])
				if obj.name in hierarchy:
					hierarchy[obj.name].append(Object.Get(element.getAttribute("bone")).name)
				else:
					hierarchy[obj.name] = [Object.Get(element.getAttribute("bone")).name]

	# gather all bone data; set their position and rotations in bone_data
	for contents in skel_xml_doc.getElementsByTagName("bones"):
		for element in contents.childNodes:
			if element.localName == "bone":
				name = element.getAttribute("name")
				obj = Object.Get(name)
				loc = None
				rot = None
				quat = None				
				for bone_xml_data in element.childNodes:
					if bone_xml_data.localName == "position":
						x = float(bone_xml_data.getAttribute("x"))
						y = float(bone_xml_data.getAttribute("y"))
						z = float(bone_xml_data.getAttribute("z"))
						loc = [x, y, z]
					elif bone_xml_data.localName == "rotation":
						# convert angle because it is in radians and Blender Quaternions are in degrees
						angle = math.degrees(float(bone_xml_data.getAttribute("angle")))
						for axis in bone_xml_data.childNodes:
							if axis.localName == "axis":
								if obj.getParent() == None:
									if fix_up_axis:
										quat = Mathutils.Euler(0.0, 0.0, 0.0).toQuat()
										rot = Mathutils.Euler(0.0, 0.0, 0.0).toMatrix()
									else:
										x = float(axis.getAttribute("x"))
										y = float(axis.getAttribute("y"))
										z = float(axis.getAttribute("z"))
										quat = Mathutils.Quaternion([x, y, z], angle)
										rot = quat.toMatrix()
									bone_root = name
								else:
									x = float(axis.getAttribute("x"))
									y = float(axis.getAttribute("y"))
									z = float(axis.getAttribute("z"))
									quat = Mathutils.Quaternion([x, y, z], angle)
									rot = quat.toMatrix()

				obj.setMatrix(rot)
				obj.setLocation(loc)
					
				bind_LocQuatInfo[name] = BoneLocQuat(loc, quat) 
				

	CreateBones(bone_root, hierarchy)
	
	
#"""
	# clean up by deleting the empties
	for obj in scene.objects:
		if obj.getType() == "Empty":
			scene.objects.unlink(obj)
#"""
def CreateSubMesh(xml_doc, name):

	global bone_index
	global has_skeleton

	scene = bpy.data.scenes.active
	vertices = []
	normals = []
	uvcoords = []
	
	mesh = bpy.data.meshes.new(name)
				
	# get the vertex data
	vertex_buffer = xml_doc.getElementsByTagName("vertexbuffer")
	for contents in vertex_buffer:
		has_positions = contents.getAttribute("positions")
		has_normals = contents.getAttribute("normals")
		for element in contents.childNodes:
			if element.localName == "vertex":
				for vertex_data in element.childNodes:
					if has_positions and vertex_data.localName == "position":
						x = float(vertex_data.getAttribute("x"))
						y = float(vertex_data.getAttribute("y"))
						z = float(vertex_data.getAttribute("z"))
						vertices.append(BlenderOrientedCoords([x, y, z]))
					elif has_normals and vertex_data.localName == "normal":
						x = float(vertex_data.getAttribute("x"))
						y = float(vertex_data.getAttribute("y"))
						z = float(vertex_data.getAttribute("z"))
						normals.append(BlenderOrientedCoords([x, y, z]))
					elif vertex_data.localName == "texcoord":
						u = float(vertex_data.getAttributeNode('u').value)
						# an OGRE/Blender format incongruence.
						# flip uv map horizontally around the 0.5 horizontal axis
						v = -float(vertex_data.getAttributeNode('v').value)+1.0
						uvcoords.append([u, v])
	# add vertex coordinates in mesh
	mesh.verts.extend(vertices)
	# add normal vectors to mesh vertices
	i = 0
	for vertex in normals:
		mesh.verts[i].no = Mathutils.Vector(vertex[0], vertex[1], vertex[2])
		i += 1

	# get the faces data
	face_data = xml_doc.getElementsByTagName("faces")[0]
	# we used [0] cause we only want the first element of the returned array.
	# there should only be 1 element in it anyway
	faces = []
	for face in face_data.childNodes:		
		if face.localName == "face":
			vertices = []
			for i in range(face.attributes.length):
				vertices.append(int(face.attributes.item(i).value))
			faces.append(vertices)
	# add faces data to mesh		
	mesh.faces.extend(faces, smooth=True)
	# add UV info if available
	if len(uvcoords) > 0:
		mesh.addUVLayer("UVLayer")
		mesh.activeUVLayer = "UVLayer"
		for face in mesh.faces:
			# uvco gets the uv coordinates of vertices associated to this face
			uvco = []
			for vertex in face.verts:
				uvco.append(Mathutils.Vector(uvcoords[vertex.index]))
			face.uv = uvco

	# add object to the current scene
	obj = scene.objects.new(mesh, name)

	# if skeleton is defined then add vertex weight info
	if has_skeleton:
		# add bone weights to vertices
		vertex_weights = xml_doc.getElementsByTagName("boneassignments")
		for contents in vertex_weights:
			for element in contents.childNodes:
				if element.localName == "vertexboneassignment":
					vertex = int(element.getAttribute("vertexindex"))
					bone = str(element.getAttribute("boneindex"))
					weight = float(element.getAttribute("weight"))

					bone_name = bone_index[bone]
					if bone_name not in mesh.getVertGroupNames():
						mesh.addVertGroup(bone_name)
					mesh.assignVertsToGroup(bone_name, [vertex], weight, Mesh.AssignModes['ADD'])
					mesh.update()
		# parent mesh to armature with deform properties
		Object.Get("Armature").makeParentDeform([obj], 0, 0)


def CreateMesh(xml_doc, name):

	# get number of sub meshes - if there are more, need to parse them separately
	if(len(xml_doc.getElementsByTagName('submesh'))>1):
		allmats = []
		for submeshes in xml_doc.getElementsByTagName('submeshes'):
			for submesh in submeshes.childNodes:
				if submesh.localName == 'submesh':
					material = str(submesh.getAttributeNode('material').value)
					CreateSubMesh(submesh, material)
					allmats.append(Object.Get(material))
					
		#TODO: move this to CreateSubMesh + parse .material file and link textures
		for matobj in allmats:
			bmat = Material.New(matobj.name)
			meshobj = matobj.getData(mesh=1)			
			meshobj.materials += [bmat]
			
			"""			
			meshobj.hasFaceUV(1)
		
			"""
				
		scene = bpy.data.scenes.active
		
		# create new mesh with 'name' and join all submeshes
		mesh = bpy.data.meshes.new(name)			
		obj = scene.objects.new(mesh, name)
		obj.join(allmats)	
		if has_skeleton:
			# parent mesh to armature with deform properties
			Object.Get("Armature").makeParentDeform([obj], 0, 0)
		
		# remove submeshes
		for ob in allmats:
			scene.objects.unlink(ob)		
		
		
	else:
		CreateSubMesh(xml_doc, name)	
	 

def AddAnimation(xml_doc):
	
	global arm
	global bind_LocQuatInfo
	global action_LocQuatInfo
	
	ParseActionSkeleton(xml_doc)
	
	#print(bind_LocQuatInfo)
	#print(action_LocQuatInfo)
	
	#return;

	# need to grab skeleton from animation (if there is any)
	# and create delta quatertion that will be added to current animation quaternions
	
	scene = bpy.data.scenes.active
	#fps = scene.getRenderingContext().framesPerSec()
	obj = Object.Get("Armature")
	scene.objects.context = [obj]
	scene.objects.active = obj
	pose = obj.getPose()

	# get the animation data
	animations = xml_doc.getElementsByTagName("animations")[0]
	for animation in animations.childNodes:
		if animation.localName == "animation":
			action = Armature.NLA.NewAction("Action")
			action.setActive(obj)
			action.fakeUser = True
			name = animation.getAttribute("name")
			length = animation.getAttribute("length")*fps
			action.setName(name)

			for tracks in animation.childNodes:
				for track in tracks.childNodes:
					if track.localName == "track":
						bone = track.getAttribute("bone")
						# get skeleton differences if flag is set
						if (fix_TLAnimation):
							animBoneDiff = GetBoneDifference(bone)
							print("getBoneDiff:" + bone)
						for keyframes in track.childNodes:
							for keyframe in keyframes.childNodes:
								if keyframe.localName == "keyframe":
									time = float(keyframe.getAttribute("time"))
									for element in keyframe.childNodes:
										if element.localName == "translate":
											x = float(element.getAttribute("x"))										
											y = float(element.getAttribute("y"))
											z = float(element.getAttribute("z"))
											loc = Mathutils.Vector(x, y, z)
											# check for flag
											if (fix_TLAnimation):
												loc[0]+=animBoneDiff.Loc[0]
												loc[1]+=animBoneDiff.Loc[1]
												loc[2]+=animBoneDiff.Loc[2]										
										elif element.localName == "rotate":
											# convert angle because it is in radians and Blender Quaternions are in degrees
											angle = math.degrees(float(element.getAttribute("angle")))
											for axis in element.childNodes:
												if axis.localName == "axis":
													x = float(axis.getAttribute("x"))
													y = float(axis.getAttribute("y"))
													z = float(axis.getAttribute("z"))
													quat = BlenderOrientedQuaternion(x, y, z, angle)													
													# check for flag
													if ((fix_TLAnimation) and (not animBoneDiff.isIdentity)):
														#quat = MultiplyQuats(animBoneDiff.Qtr, Mathutils.Quaternion([x, y, z], -angle))
														#quat = BlenderOrientedQuaternion(quat.axis[0], quat.axis[1], quat.axis[2], quat.angle)
														#quat = MultiplyQuats(animBoneDiff.Qtr, BlenderOrientedQuaternion(x, y, z, angle).normalize())
														#quat = MultiplyQuats(BlenderOrientedQuaternion(x, y, z, -angle), animBoneDiff.Qtr)
														# dstKeyFrame->setRotation(deltaTransform.rotate * srcKeyFrame->getRotation());
														#quat = MultiplyQuats(animBoneDiff.Qtr, BlenderOrientedQuaternion(quat.axis[0], quat.axis[1], quat.axis[2], -quat.angle))
														#quat = MultiplyQuats(Mathutils.Quaternion(animBoneDiff.Qtr.axis[0], animBoneDiff.Qtr.axis[1], animBoneDiff.Qtr.axis[2], -animBoneDiff.Qtr.angle), quat)
														#quat = MultiplyQuats(BlenderOrientedQuaternion(-animBoneDiff.Qtr.axis[1], animBoneDiff.Qtr.axis[0], animBoneDiff.Qtr.axis[2], -animBoneDiff.Qtr.angle), quat)
														#print(animBoneDiff.Qtr.x)
														
														quat = MultiplyQuats(BlenderOrientedQuaternion(animBoneDiff.Qtr.axis[0], animBoneDiff.Qtr.axis[1], animBoneDiff.Qtr.axis[2], -animBoneDiff.Qtr.angle), quat)
														
														# TODO: this looks wrong, those values are already from quaternion														

									if pose.bones[bone] != None:
										pose.bones[bone].quat = quat
										pose.bones[bone].loc = loc
										pose.bones[bone].insertKey(obj, int(fps*time), [Object.Pose.ROT, Object.Pose.LOC], True)
									else:
										pass
									pose.update()




def ImportOgre(path):

	global has_skeleton

	Window.WaitCursor(1)
	folder = os.path.split(path)[0]
	name = "mesh"
	files = []

	if not DEBUG:	
		# convert MESH and SKELETON file to MESH.XML and SKELETON.XML respectively
		for filename in os.listdir(folder):
			# we're going to do string comparisons. assume lower case to simplify code
			filename = os.path.join(folder, filename.lower())
			# process .mesh and .skeleton files while skipping .xml files
			if ((".skeleton" in filename) or (".mesh" in filename)) and (".xml" not in filename):
				os.system('%s "%s"' % (ogreXMLConverter, filename))

	# get all the filenames in the chosen directory, put in list and sort it
	for filename in os.listdir(folder):
		# we're going to do string comparisons. assume lower case to simplify code
		filename = filename.lower()
		# process .mesh and .skeleton files while skipping .xml files
		if ".skeleton.xml" in filename:
			files.append(os.path.join(folder, filename))
		elif ".mesh.xml" in filename:
			# get the name of the MESH file without extension. Use this base name to name our imported object
			name = filename.split('.')[0]
			# put MESH file on top of the file list
			files.insert(0, os.path.join(folder, filename))

	# now that we have a list of files, process them
	filename = files[0]
	# import the mesh
	if (".mesh" in filename):
		mesh_data = OpenFile(filename)
		if mesh_data != "None":
			CreateSkeleton(mesh_data, folder, name)
			CreateMesh(mesh_data, name)
			if not DEBUG:
				# cleanup by deleting the XML file we created
				os.unlink("%s" % filename)
	
	if has_skeleton:
		# process all the files again. this time process skeleton files which must be processed after mesh file
		# the armature must already be created (when mesh was processed) prior to this.
		for filename in files:
			# import the skeleton file
			if (".skeleton" in filename):
				skeleton_data = OpenFile(filename)
				if skeleton_data != "None":
					AddAnimation(skeleton_data)
					if not DEBUG:
						# cleanup by deleting the XML file we created
						os.unlink("%s" % filename)

	Window.WaitCursor(0)
	Window.RedrawAll()



print "\n\n\n"
Window.FileSelector(ImportOgre, "Import")
